{$F+}
{
  -- Programmer : Antoun Kanawati.
  -- Date       : November 23, 1988
  -- Version    : 1.0, revised for Turbo Pascal 5.0.
  --              original written under Turbo Pascal 3.01A.
  -- Unit       : inputil, a small set of input utilities.
  -- History    :
  --              The original source of this unit was a TP3 include file.
  --              The TP3 version did not define or use console_in/console_out
  --              because the TRM (and CON) special files provided their 
  --              services.  The TP3 version did not check the validity of 
  --              inputs either.  The TP3 version also got confused with 
  --              extended codes, because TP3 coded them as <ESC><CODE> 
  --              instead of the usual <NUL><CODE>; thus when <ESC> followed
  --              by another key were hit before calling inkey, the function
  --              thought it had an extended code, and behaved accordingly.
  -- Portability:
  --              For use with TP4, the "USES crt;" statement in the 
  --              implementation section must be moved up to the interface
  --              section.
  -- Interface  :
  --              console_in, console_out
  --                Files to access console whether I/O has been redirected 
  --                or not (similar to 3.0's trm, and con, but are one way).
  --
  --              inkey
  --                function : [keyboard] -> integer
  --                  reads keyboard,
  --                  returns ascii code of character read, when a regular
  --                    key is pressed,
  --                  returns a negative code for extended keys, such as
  --                    function keys, and cursor keypad.
  --
  --              say_yes
  --                function : message-string [keyboard] -> boolean
  --                  prints message, returns TRUE if user presses Y (or y),
  --                  FALSE if user presses N (or n).
  --
  --              get_string
  --                function : char-set integer integer boolean -> string
  --                  reads a string from the keyboard filtering it thru
  --                  char-set, and limiting its length between the two
  --                  integers.  Returns when user presses ENTER, and the
  --                  string is acceptable.  If the boolean is true, the
  --                  ENTER is echoed as a newline.
  --                  i.e.: char-set : set of allowable characters, only
  --                        those are accepted for input, others are ignored.
  --                        1st integer : minimum length of string.
  --                        2nd integer : maximum length of string.
  --                        boolean     : TRUE echo newline, FALSE do not
  --                                      echo newline.
  --                        Special characters :
  --                          Backspace, carriage return, Escape.
  --                          Backspace : Delete character before cursor.
  --                          carriage return : Terminate input.
  --                          Escape : Restart input.
  --                          All other control characters are NOT allowed,
  --                          regardless of what the input character set
  --                          contains.  An error is signalled if the specs
  --                          are not proper.
  --                        Limitations :
  --                          Works on a single line only.  If specs
  --                          do not fit on one line, behaviour is rather
  --                          messy.
  --
  --              read_int
  --                procedure(var n : integer; Min, Max : integer);
  --                  reads a numeric string from keyboard, allowing only
  --                  digits, and '-'.  If the numeric string converts to
  --                  an integer within Min..Max, the converted value is
  --                  stored in n, and the procedure returns.
  --                  If the input is a null string, n is not changed,
  --                  and the procedure returns.
  --                  Otherwise, the procedure insists on getting a value.
  --                  An error is signalled when limits are not proper.
  --
  --              read_real
  --                procedure (var n : real; Min, Max : real);
  --                  performs same as read_int, but with real's.
}

{$n+} {-- required for read_int --}
unit inputil;

interface


  var
    console_in,
    console_out : text;

  type
    Msg_String = string[70];
    Char_Set   = set of char;
    Big_String = string[255];

function inkey : integer;
function say_yes(msg : msg_string) : boolean;
function get_string(valid_set  : char_set;
                    min_length,
                    max_length : integer;
                    new_line   : boolean) : big_string;
procedure read_int(var n : integer; min, max : integer);
procedure read_real(var n : real; Min, Max : real);
{------------------------------------------------------------------------}

implementation

  uses crt;

  const
    _BS        = #8;
    _CR        = #13;
    _ESC       = #27;
    _FILLER    = '_';

{------------------------------------------------------------------------}
  function inkey : integer;

    var
      c : char;

    begin
      c := readkey;
      if ((c = #0) and keypressed)
        then
          begin
            c := readkey;
            inkey := -ord(c)
          end
        else
          inkey := ord(c)
    end;
{------------------------------------------------------------------------}
  function Say_Yes( Msg : Msg_String) : boolean;

    var
      c : integer;

    begin
      write(console_out, Msg, '<Y/N> : ', _FILLER, _BS);

      repeat
        repeat
          c := inkey
        until (c > 0);
        c := ord( upcase(chr(c)) )
      until (chr(c) in ['Y','N']);

      if (chr(c) = 'Y')
        then
          writeln(console_out, 'YES')
        else
          writeln(console_out, 'NO');

      Say_Yes := chr(c) = 'Y'
    end;
{------------------------------------------------------------------------}
  var
    Buffer : Big_String; {-- Minimize stack use. --}
  function get_string( Valid_Set   : Char_Set;
                       Min_Length,
                       Max_Length  : integer;
                       New_Line    : boolean  ) : Big_String;

    var
      c, i   : integer;
      Done   : boolean;

    begin
      Valid_Set := Valid_Set - [#0 .. #31, #127];

      if ((Valid_Set = []) and (Min_Length > 0)) then
        begin
          writeln(console_out, 'get_string: Invalid specs.  ',
                  'Allowable input character-set is empty,',
                  'and minimum input-length is greater than zero.');
          halt(1)
        end;

      if ((Min_Length > Max_Length) or (Min_Length < 0) or (Max_Length < 0))
        then
          begin
            writeln(console_out, 'get_string: Invalid input length limits.  ',
                    'Min_Length, and Max_Length must be >= 0, ',
                    'and must be Min_Length <= Max_Length.');
            writeln(console_out, 'Input was <Min, Max> : ', Min_Length, ', ',
                                                            Max_Length);

            halt(1)
          end;

      Done      := false;
      Buffer    := '';

      for i := 1 to Max_Length do
        write(console_out, _FILLER );

      for i := 1 to Max_Length do
        write(console_out, _BS);

      repeat
        repeat
          c := inkey
        until (c > 0);

        case chr(c) of
          _ESC : if (length(Buffer) > 0)
                   then
                     begin
                       for i := 1 to length(Buffer) do
                         write(console_out, _BS, _FILLER, _BS);
                       Buffer := ''
                     end;

          _BS  : if (length(Buffer) > 0)
                   then
                     begin
                       delete(Buffer,length(Buffer),1);
                       write(console_out, _BS, _FILLER, _BS)
                     end;

          _CR  : Done := (length(Buffer) >= Min_Length)

          else   if (chr(c) in Valid_Set)
                   then
                     if (length(Buffer) < Max_Length)
                       then
                         begin
                           Buffer := Buffer + chr(c);
                           write(console_out, chr(c))
                         end
        end
      until Done;

      if New_Line then
        writeln(console_out);

      Get_String := Buffer
    end;
{------------------------------------------------------------------------}
  procedure read_int(var n : integer; Min, Max : integer);

    var
      s       : Big_String;
      v       : comp;
      errpos  : integer;
      x,y     : integer;

    begin
      x := wherex;
      y := wherey;

      if (Min > Max) then
        begin
          writeln(console_out, 'read_int: improper Min/Max limits.');
          writeln(console_out, 'Min = ', Min, ', Max = ', Max);
          halt(1)
        end;

      repeat
        gotoxy(x,y);
        s := Get_String( ['0'..'9','-'], 0, 6, false);
        if (s = '')
          then
            begin
              gotoxy(x,y);
              writeln(console_out, 'NO CHANGE.');
              exit
            end;

        val(s, v, errpos)
      until ((errpos = 0) and (v >= Min) and (v <= Max));

      n := trunc(v);
      writeln(console_out)
    end;
{------------------------------------------------------------------------}
  procedure read_real(var n : real; Min, Max : real);

    var
      s       : Big_String;
      v       : real;
      errpos  : integer;
      x,y     : integer;

    begin
      x := wherex;
      y := wherey;

      if (Min > Max) then
        begin
          writeln(console_out, 'read_real: improper Min/Max limits.');
          writeln(console_out, 'Min = ', Min, ', Max = ', Max);
          halt(1)
        end;

      repeat
        gotoxy(x,y);
        s := Get_String( ['0'..'9','-','.','e','E'], 0, 15, false);
        if (s = '')
          then
            begin
              gotoxy(x,y);
              writeln(console_out,'NO CHANGE.');
              exit
            end;

        val(s, v, errpos)
      until ((errpos = 0) and (v >= Min) and (v <= Max));

      n := v;
      writeln(console_out)
    end;

{------------------------------------------------------------------------}
begin
  assign(console_in, 'CON');
  assign(console_out, 'CON');
  reset(console_in);
  rewrite(console_out)

end.

